import itertools
from typing import List

from HABApp.core.logger import HABAppError
from HABApp.openhab.definitions.helpers.log_table import Table
from ._log import log_cfg as log
from .thing_config import ThingConfigChanger


def show_config_overview(cfgs: List[ThingConfigChanger], all_params):
    t = Table(heading='Current configuration')
    c_p = t.add_column('Parameter')
    c_s = [t.add_column(c.uid, alias=':'.join(c.uid.split(':')[-2:]) if ':' in c.uid else None) for c in cfgs]

    for p in itertools.chain(
            sorted(filter(lambda x: isinstance(x, int), all_params)),
            sorted(filter(lambda x: not isinstance(x, int), all_params))):
        c_p.add(str(p))
        for col, cfg in zip(c_s, cfgs):
            v = cfg.get(p, '')
            col.add(str(v))
    for line in t.get_lines():
        log.info(line)


async def update_thing_cfg(target_cfg, things, test: bool):

    cfgs = [ThingConfigChanger.from_dict(k['UID'], k.get('configuration', {})) for k in things]
    cur_vals = tuple(k.get_dict(filter=True) for k in cfgs)
    all_params = set()
    for k in cur_vals:
        all_params.update(k.keys())

    # Show overview
    if test:
        show_config_overview(cfgs, all_params)

    errs = HABAppError(log)
    skip_cfg = set()
    for param, value in target_cfg.items():
        for c in cfgs:
            try:
                c[param] = value
            except Exception as e:
                errs.add_exception(e)
                skip_cfg.add(c)

    if errs:
        errs.dump()
        if not test:
            show_config_overview(cfgs, all_params)

    # update the thing configs. If nothing has changed this will do nothing
    for c in cfgs:
        # do not create partially correct configs
        if c in skip_cfg:
            log.warning(f'Skip configuration for {c.uid} because of errors!')
            continue

        if test:
            if c.new:
                log.info(f'Would set {c.get_dict(True, True)} for {c.uid}')
            else:
                log.info(f'Nothing changed for {c.uid}')
        else:
            await c.update_thing_cfg()

    return None



import logging
from pathlib import Path

from watchdog.events import FileSystemEvent
from watchdog.events import EVENT_TYPE_CLOSED as WD_EVENT_TYPE_CLOSED
from watchdog.events import EVENT_TYPE_OPENED as WD_EVENT_TYPE_OPENED


log = logging.getLogger('HABApp.file.events')
log.setLevel(logging.INFO)


class EventFilterBase:
    def notify(self, path: str) -> bool:
        raise NotImplementedError()


class FileEndingFilter(EventFilterBase):
    def __init__(self, ending: str):
        self.ending: str = ending

    def notify(self, path: str) -> bool:
        return path.endswith(self.ending)

    def __repr__(self):
        return f'<{self.__class__.__name__} ending: {self.ending}>'


class FileSystemEventHandler:
    def __init__(self, folder: Path, filter: EventFilterBase, watch_subfolders: bool = False):
        assert isinstance(folder, Path), type(folder)
        assert watch_subfolders is True or watch_subfolders is False

        self.folder: Path = folder
        self.watch_subfolders: bool = watch_subfolders

        self.filter: EventFilterBase = filter

    def dispatch(self, event: FileSystemEvent):
        log.debug(event)

        # we don't process directory events
        if event.is_directory:
            return None

        # we don't process open and close events
        if (e_type := event.event_type) == WD_EVENT_TYPE_OPENED or e_type == WD_EVENT_TYPE_CLOSED:
            return None

        src = event.src_path
        if self.filter.notify(src):
            self.file_changed(src)

        # moved events have a dst, so we process it, too
        if hasattr(event, 'dest_path'):
            dst = event.dest_path
            if self.filter.notify(dst):
                self.file_changed(dst)
        return None

    def file_changed(self, dst: str):
        raise NotImplementedError()



from datetime import timedelta
from unittest.mock import Mock

import pytest

from HABApp import __cmd_args__


def test_get_uptime():
    __cmd_args__.get_uptime()


@pytest.mark.parametrize('arg_name', ('-wos', '--wait_os_uptime'))
def test_cmd_wait_uptime(monkeypatch, arg_name):
    td = timedelta(days=110 * 365)

    m = Mock()
    monkeypatch.setattr(__cmd_args__.time, 'sleep', m)
    monkeypatch.setattr(__cmd_args__, 'find_config_folder', lambda x: x)

    __cmd_args__.parse_args(['-w', f'{int(td.total_seconds())}'])

    # get first func call
    args = m.call_args[0]
    waited = timedelta(seconds=args[0])
    assert waited.days > 100 * 365



from __future__ import annotations

import datetime
from typing import Any

from HABApp.core.asyncio import run_coro_from_thread
from HABApp.core.internals import ItemRegistryItem
from .func_async import async_create_item, async_get_item, \
    async_get_thing, async_set_thing_enabled, \
    async_set_metadata, async_remove_metadata, \
    async_remove_item, async_item_exists, async_get_persistence_data, async_set_persistence_data, \
    async_get_persistence_services, async_get_link, async_create_link, async_remove_link
from HABApp.openhab import definitions
from HABApp.openhab.definitions.helpers import OpenhabPersistenceData
from HABApp.openhab.definitions.rest import ItemResp, ItemChannelLinkResp


# ----------------------------------------------------------------------------------------------------------------------
# /items
# ----------------------------------------------------------------------------------------------------------------------
# def get_items():
#     """Request all items from the openHAB Rest API. Don't use this
#     """
#     return run_coro_from_thread(async_get_items())


def get_item(item: str | ItemRegistryItem) -> ItemResp | None:
    """Return the complete openHAB item definition

    :param item: name of the item or item
    :return: openHAB item
    """
    assert isinstance(item, (str, ItemRegistryItem)), type(item)

    return run_coro_from_thread(async_get_item(item), calling=get_item)


def item_exists(item: str | ItemRegistryItem):
    """
    Check if an item exists in the openHAB item registry

    :param item: name of the item or item
    :return: True if item was found
    """
    assert isinstance(item, (str, ItemRegistryItem)), type(item)

    return run_coro_from_thread(async_item_exists(item), calling=item_exists)


def remove_item(item: str | ItemRegistryItem):
    """
    Removes an item from the openHAB item registry

    :param item: name
    :return: True if item was found and removed
    """
    assert isinstance(item, (str, ItemRegistryItem)), type(item)

    return run_coro_from_thread(async_remove_item(item), calling=remove_item)


def create_item(item_type: str, name: str,
                label: str | None = None, category: str | None = None,
                tags: list[str] | None = None, groups: list[str] | None = None,
                group_type: str | None = None,
                group_function: str | None = None,
                group_function_params: list[str] | None = None) -> bool:
    """Creates a new item in the openHAB item registry or updates an existing one

    :param item_type: item type
    :param name: item name
    :param label: item label
    :param category: item category
    :param tags: item tags
    :param groups: in which groups is the item
    :param group_type: what kind of group is it
    :param group_function: group state aggregation function
    :param group_function_params: params for group state aggregation
    :return: True if item was created/updated
    """

    def validate(_in):
        assert isinstance(_in, str), type(_in)

    # limit values to special entries and validate parameters
    if ':' in item_type:
        __type, __unit = item_type.split(':')
        assert __unit in definitions.ITEM_DIMENSIONS, \
            f'{__unit} is not a valid openHAB unit: {", ".join(definitions.ITEM_DIMENSIONS)}'
        assert __type in definitions.ITEM_TYPES, \
            f'{__type} is not a valid openHAB type: {", ".join(definitions.ITEM_TYPES)}'
    else:
        assert item_type in definitions.ITEM_TYPES, \
            f'{item_type} is not an openHAB type: {", ".join(definitions.ITEM_TYPES)}'
    assert isinstance(name, str), type(name)
    assert isinstance(label, str) or label is None, type(label)
    assert isinstance(category, str) or category is None, type(category)
    if tags:
        map(validate, tags)
    if groups:
        map(validate, groups)
    if group_type:
        assert isinstance(group_type, str), type(group_type)
    if group_function:
        assert isinstance(group_function, str), type(group_function)
    if group_function_params:
        map(validate, group_function_params)

    if group_type or group_function or group_function_params:
        assert item_type == 'Group', f'Item type must be "Group"! Is: {item_type}'

        if group_function:
            assert group_function in definitions.GROUP_ITEM_FUNCTIONS, \
                f'{item_type} is not a group function: {", ".join(definitions.GROUP_ITEM_FUNCTIONS)}'

    return run_coro_from_thread(
        async_create_item(
            item_type, name,
            label=label, category=category, tags=tags, groups=groups,
            group_type=group_type, group_function=group_function, group_function_params=group_function_params
        ),
        calling=create_item
    )


def set_metadata(item: str | ItemRegistryItem, namespace: str, value: str, config: dict):
    """
    Add/set metadata to an item

    :param item: name of the item or item
    :param namespace: namespace, e.g. ``stateDescription``
    :param value: value
    :param config: configuration e.g. ``{"options": "A,B,C"}``
    :return: True if metadata was successfully created/updated
    """
    assert isinstance(item, (str, ItemRegistryItem)), type(item)
    assert isinstance(namespace, str), type(namespace)
    assert isinstance(value, str), type(value)
    assert isinstance(config, dict), type(config)

    return run_coro_from_thread(
        async_set_metadata(item=item, namespace=namespace, value=value, config=config), calling=set_metadata
    )


def remove_metadata(item: str | ItemRegistryItem, namespace: str):
    """
    Remove metadata from an item

    :param item: name of the item or item
    :param namespace: namespace
    :return: True if metadata was successfully removed
    """
    assert isinstance(item, (str, ItemRegistryItem)), type(item)
    assert isinstance(namespace, str), type(namespace)

    return run_coro_from_thread(async_remove_metadata(item=item, namespace=namespace), calling=remove_metadata)


# ----------------------------------------------------------------------------------------------------------------------
# /things
# ----------------------------------------------------------------------------------------------------------------------
def get_thing(thing: str | ItemRegistryItem):
    """ Return the complete openHAB thing definition

    :param thing: name of the thing or the item
    :return: openHAB thing
    """
    assert isinstance(thing, (str, ItemRegistryItem)), type(thing)

    return run_coro_from_thread(async_get_thing(thing), calling=get_thing)


def set_thing_enabled(thing: str | ItemRegistryItem, enabled: bool = True):
    """
    Enable/disable a thing

    :param thing: name of the thing or the thing object
    :param enabled: True to enable thing, False to disable thing
    """
    assert isinstance(thing, (str, ItemRegistryItem)), type(thing)

    return run_coro_from_thread(async_set_thing_enabled(thing, enabled), calling=set_thing_enabled)


# ----------------------------------------------------------------------------------------------------------------------
# /persistence
# ----------------------------------------------------------------------------------------------------------------------
def get_persistence_services():
    """ Return all available persistence services
    """
    return run_coro_from_thread(async_get_persistence_services(), calling=get_persistence_services)


def get_persistence_data(item: str | ItemRegistryItem, persistence: str | None,
                         start_time: datetime.datetime | None,
                         end_time: datetime.datetime | None) -> OpenhabPersistenceData:
    """Query historical data from the openHAB persistence service

    :param item: name of the persistent item
    :param persistence: name of the persistence service (e.g. ``rrd4j``, ``mapdb``). If not set default will be used
    :param start_time: return only items which are newer than this
    :param end_time: return only items which are older than this
    :return: last stored data from persistency service
    """
    assert isinstance(item, (str, ItemRegistryItem)), type(item)
    assert isinstance(persistence, str) or persistence is None, persistence
    assert isinstance(start_time, datetime.datetime) or start_time is None, start_time
    assert isinstance(end_time, datetime.datetime) or end_time is None, end_time

    ret = run_coro_from_thread(
        async_get_persistence_data(item=item, persistence=persistence, start_time=start_time, end_time=end_time),
        calling=get_persistence_data
    )
    return OpenhabPersistenceData.from_resp(ret)


def set_persistence_data(item: str | ItemRegistryItem, persistence: str | None, time: datetime.datetime, state: Any):
    """Set a measurement for a item in the persistence serivce

    :param item_name: name of the persistent item
    :param persistence: name of the persistence service (e.g. ``rrd4j``, ``mapdb``). If not set default will be used
    :param time: time of measurement
    :param state: state which will be set
    :return: True if data was stored in persistency service
    """
    assert isinstance(item, (str, ItemRegistryItem)), type(item)
    assert isinstance(persistence, str) or persistence is None, persistence
    assert isinstance(time, datetime.datetime), time

    return run_coro_from_thread(
        async_set_persistence_data(item=item, persistence=persistence, time=time, state=state),
        calling=set_persistence_data
    )


# ---------------------------------------------------------------------------------------------------------------------
# Link handling is experimental
# ---------------------------------------------------------------------------------------------------------------------
def get_link(item: str | ItemRegistryItem, channel: str) -> ItemChannelLinkResp:
    """ returns the link between an item and a (things) channel

    :param item: name of the item or item
    :param channel: uid of the (things) channel (usually something like AAAA:BBBBB:CCCCC:DDDD:0#SOME_NAME)
    """
    assert isinstance(item, (str, ItemRegistryItem)), type(item)
    assert isinstance(channel, str), type(channel)

    return run_coro_from_thread(async_get_link(item, channel), calling=get_link)


def create_link(
        item: str | ItemRegistryItem, channel: str, configuration: dict[str, Any] | None = None) -> bool:
    """creates a link between an item and a (things) channel

    :param item: name of the item or item
    :param channel: uid of the (things) channel (usually something like AAAA:BBBBB:CCCCC:DDDD:0#SOME_NAME)
    :param configuration: optional configuration for the channel
    :return: True on successful creation, otherwise False
    """
    assert isinstance(item, (str, ItemRegistryItem)), type(item)
    assert isinstance(channel, str), type(channel)
    assert isinstance(configuration, dict), type(configuration)

    return run_coro_from_thread(async_create_link(item, channel, configuration), calling=create_link)


def remove_link(item: str | ItemRegistryItem, channel: str) -> bool:
    """ removes a link between a (things) channel and an item

    :param item: name of the item or item
    :param channel: uid of the (things) channel (usually something like AAAA:BBBBB:CCCCC:DDDD:0#SOME_NAME)
    :return: True on successful removal, otherwise False
    """
    assert isinstance(item, (str, ItemRegistryItem)), type(item)
    assert isinstance(channel, str), type(channel)

    return run_coro_from_thread(async_remove_link(item, channel), calling=remove_link)



from __future__ import annotations

from asyncio import sleep, Task, create_task, CancelledError

from HABApp.core.connections import BaseConnection, BaseConnectionPlugin


class WaitBetweenConnects:
    wait_max = 600

    def __init__(self):
        self.wait_time: int = 0
        self.task: Task | None = None

    def reset_wait(self):
        self.wait_time = 0

    async def wait(self):
        wait = self.wait_time
        wait = wait * 2 if wait <= 16 else wait * 1.5
        wait = max(1, min(wait, self.wait_max))

        self.wait_time = wait

        try:
            self.task = create_task(sleep(self.wait_time))
            await self.task
        except CancelledError:
            pass
        finally:
            self.task = None

    def cancel(self):
        if task := self.task:
            task.cancel()


class AutoReconnectPlugin(BaseConnectionPlugin):
    _DEFAULT_PRIORITY = 110_000

    def __init__(self, name: str | None = None):
        super().__init__(name)
        self.waiter = WaitBetweenConnects()

    def on_application_shutdown(self):
        self.waiter.cancel()

    async def on_online(self):
        self.waiter.reset_wait()

    async def on_offline(self, connection: BaseConnection):
        if connection.is_shutdown:
            return None

        if connection.has_errors:
            await self.waiter.wait()
            connection.clear_error()



import logging
from typing import Generic, TypeVar, List, Union, Type

from pendulum import DateTime

from HABApp.core.asyncio import run_func_from_async
from .base_item_watch import BaseWatch, ItemNoChangeWatch, ItemNoUpdateWatch

log = logging.getLogger('HABApp')

WATCH_OBJ = TypeVar("WATCH_OBJ", bound=BaseWatch)


class ItemTimes(Generic[WATCH_OBJ]):
    WATCH: Union[Type[ItemNoUpdateWatch], Type[ItemNoChangeWatch]]

    def __init__(self, name: str, dt: DateTime):
        self.name: str = name
        self.dt: DateTime = dt
        self.tasks: List[WATCH_OBJ] = []

    def set(self, dt: DateTime, events=True):
        self.dt = dt
        if not self.tasks:
            return

        if events:
            run_func_from_async(self.__async_schedule_events)
        return None

    def add_watch(self, secs: Union[int, float]) -> WATCH_OBJ:
        # don't add the watch two times
        for t in self.tasks:
            if not t.fut.is_canceled and t.fut.secs == secs:
                log.warning(f'Watcher {self.WATCH.__name__} ({t.fut.secs}s) for {self.name} has already been created')
                return t

        w = self.WATCH(self.name, secs)
        self.tasks.append(w)
        log.debug(f'Added {self.WATCH.__name__} ({w.fut.secs}s) for {self.name}')
        return w

    def __async_schedule_events(self):
        canceled = []
        for t in self.tasks:
            if t.fut.is_canceled:
                canceled.append(t)
            else:
                t.fut.reset()

        # remove canceled tasks
        if canceled:
            for c in canceled:
                self.tasks.remove(c)
                log.debug(f'Removed {self.WATCH.__name__} ({c.fut.secs}s) for {self.name}')
        return None


class UpdatedTime(ItemTimes[ItemNoUpdateWatch]):
    WATCH = ItemNoUpdateWatch


class ChangedTime(ItemTimes[ItemNoChangeWatch]):
    WATCH = ItemNoChangeWatch



from typing import TYPE_CHECKING, TypeVar, Optional, Any

import HABApp
from HABApp.core.internals import AutoContextBoundObj


class BaseMode(AutoContextBoundObj):

    def __init__(self, name: str):
        super().__init__()
        self.name: str = name

        self.__mode_lower_prio: Optional[BaseMode] = None

        self.parent: HABApp.util.multimode.MultiModeItem

        # Otherwise the missing assignment shows an error
        if TYPE_CHECKING:
            self.parent = HABApp.util.multimode.MultiModeItem('TYPE_CHECKING')
        return

    def _set_mode_lower_prio(self, mode_lower_prio):
        assert isinstance(mode_lower_prio, BaseMode) or mode_lower_prio is None, type(mode_lower_prio)
        self.__lower_priority_mode = mode_lower_prio

    def calculate_value(self, lower_prio_value: Any) -> Any:
        raise NotImplementedError()

    def cancel(self):
        """Remove the mode from the parent ``MultiModeItem`` and stop processing it
        """
        self._ctx_unlink()
        self.parent.remove_mode(self.name)
        self.parent = None


HINT_BASE_MODE = TypeVar('HINT_BASE_MODE', bound=BaseMode)



import datetime
from typing import Any, FrozenSet, Mapping, NamedTuple, Optional, TypeVar, Type

from immutables import Map

from HABApp.core.const import MISSING
from HABApp.core.items import BaseValueItem
from HABApp.core.lib.funcs import compare as _compare
from HABApp.openhab.interface_sync import get_persistence_data, post_update, send_command


class MetaData(NamedTuple):
    value: str
    config: Mapping[str, Any] = Map()


class OpenhabItem(BaseValueItem):
    """Base class for items which exists in OpenHAB.

    :ivar str name:
    :ivar Any value:

    :ivar Optional[str] label:
    :ivar FrozenSet[str] tags:
    :ivar FrozenSet[str] groups:
    :ivar Mapping[str, MetaData] metadata:
    """

    def __init__(self, name: str, initial_value: Any = None,
                 label: Optional[str] = None, tags: FrozenSet[str] = frozenset(), groups: FrozenSet[str] = frozenset(),
                 metadata: Mapping[str, MetaData] = Map()):
        super().__init__(name, initial_value)
        self.label: Optional[str] = label
        self.tags: FrozenSet[str] = tags
        self.groups: FrozenSet[str] = groups
        self.metadata: Mapping[str, MetaData] = metadata

    @classmethod
    def from_oh(cls, name: str, value=None,
                label: Optional[str] = None, tags: FrozenSet[str] = frozenset(), groups: FrozenSet[str] = frozenset(),
                metadata: Mapping[str, MetaData] = Map()):
        if value is not None:
            value = cls._state_from_oh_str(value)
        return cls(name, value, label=label, tags=tags, groups=groups, metadata=metadata)

    @staticmethod
    def _state_from_oh_str(state: str):
        raise NotImplementedError()

    def oh_send_command(self, value: Any = MISSING):
        """Send a command to the openHAB item

        :param value: (optional) value to be sent. If not specified the current item value will be used.
        """
        send_command(self.name, self.value if value is MISSING else value)

    def oh_post_update(self, value: Any = MISSING):
        """Post an update to the openHAB item

        :param value: (optional) value to be posted. If not specified the current item value will be used.
        """
        post_update(self.name, self.value if value is MISSING else value)

    def oh_post_update_if(self, new_value, *, equal=MISSING, eq=MISSING, not_equal=MISSING, ne=MISSING,
                          lower_than=MISSING, lt=MISSING, lower_equal=MISSING, le=MISSING,
                          greater_than=MISSING, gt=MISSING, greater_equal=MISSING, ge=MISSING,
                          is_=MISSING, is_not=MISSING) -> bool:
        """
        Post a value depending on the current state of the item. If one of the comparisons is true the new state
        will be posted.

        :param new_value: new value to post
        :param equal: item state has to be equal to the passed value
        :param eq: item state has to be equal to the passed value
        :param not_equal: item state has to be not equal to the passed value
        :param ne: item state has to be not equal to the passed value
        :param lower_than: item state has to be lower than the passed value
        :param lt: item state has to be lower than the passed value
        :param lower_equal: item state has to be lower equal the passed value
        :param le: item state has to be lower equal the passed value
        :param greater_than: item state has to be greater than the passed value
        :param gt: item state has to be greater than the passed value
        :param greater_equal: item state has to be greater equal the passed value
        :param ge: item state has to be greater equal the passed value
        :param is_: item state has to be the same object as the passt value (e.g. None)
        :param is_not: item state has to be not the same object as the passt value (e.g. None)

        :return: `True` if the new value was posted else `False`
        """

        if _compare(self.value, equal=equal, eq=eq, not_equal=not_equal, ne=ne,
                    lower_than=lower_than, lt=lt, lower_equal=lower_equal, le=le,
                    greater_than=greater_than, gt=gt, greater_equal=greater_equal, ge=ge, is_=is_, is_not=is_not):
            post_update(self.name, new_value)
            return True
        return False

    def get_persistence_data(self, persistence: Optional[str] = None,
                             start_time: Optional[datetime.datetime] = None,
                             end_time: Optional[datetime.datetime] = None):
        """Query historical data from the OpenHAB persistence service

        :param persistence: name of the persistence service (e.g. ``rrd4j``, ``mapdb``). If not set default will be used
        :param start_time: return only items which are newer than this
        :param end_time: return only items which are older than this
        """

        return get_persistence_data(
            self.name, persistence, start_time, end_time
        )


HINT_OPENHAB_ITEM = TypeVar('HINT_OPENHAB_ITEM', bound=OpenhabItem)
HINT_TYPE_OPENHAB_ITEM = Type[HINT_OPENHAB_ITEM]



from colorama import just_fix_windows_console, deinit, Fore

from HABApp.__version__ import __version__


def show_screen():
    text = r"""
  _   _    _    ____    _
 | | | |  / \  | __ )  / \   _ __  _ __
 | |_| | / _ \ |  _ \ / _ \ | '_ \| '_ \
 |  _  |/ ___ \| |_) / ___ \| |_) | |_) |
 |_| |_/_/   \_|____/_/   \_| .__/| .__/
                            |_|   |_|
"""

    red = r"""
    _
   / \   _ __  _ __
  / _ \ | '_ \| '_ \
 / ___ \| |_) | |_) |
/_/   \_| .__/| .__/
        |_|   |_|
"""

    just_fix_windows_console()

    for line_text, line_red in zip(text.strip('\n\r').splitlines(), red.strip('\n\r').splitlines()):
        red = line_red.strip()
        assert line_text.endswith(red)
        print(f'{line_text[:-len(red)]:s}{Fore.RED}{line_text[-len(red):]}{Fore.RESET}')

    # Version nr
    print(f'{" " * 40}{__version__}')

    deinit()



from typing import Dict, List
from typing import Optional, Any

from msgspec import Struct, field

from HABApp.openhab.definitions import ThingStatusEnum, ThingStatusDetailEnum


class ChannelResp(Struct, kw_only=True):
    # ChannelDTO
    # https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core.thing/src/main/java/org/openhab/core/thing/dto/ChannelDTO.java

    uid: str
    id: str
    channel_type: str = field(name='channelTypeUID')
    item_type: Optional[str] = field(default=None, name='itemType')
    kind: str
    label: str = ''
    description: str = ''
    default_tags: List[str] = field(default=list, name='defaultTags')
    properties: Dict[str, Any] = {}
    configuration: Dict[str, Any] = {}
    auto_update_policy: str = field(default='', name='autoUpdatePolicy')

    # EnrichedChannelDTO
    # https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core.io.rest.core/src/main/java/org/openhab/core/io/rest/core/thing/EnrichedChannelDTO.java
    linked_items: List[str] = field(name='linkedItems')


# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core.thing/src/main/java/org/openhab/core/thing/ThingStatusInfo.java
class ThingStatusResp(Struct):
    status: ThingStatusEnum
    detail: ThingStatusDetailEnum = field(name='statusDetail')
    description: Optional[str] = None


# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core.thing/src/main/java/org/openhab/core/thing/firmware/dto/FirmwareStatusDTO.java
class FirmwareStatusResp(Struct):
    status: str
    updatable_version: Optional[str] = field(default=None, name='updatableVersion')


class ThingResp(Struct, kw_only=True):
    # AbstractThingDTO
    # https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core.thing/src/main/java/org/openhab/core/thing/dto/AbstractThingDTO.java
    label: str = ''
    bridge_uid: Optional[str] = field(default=None, name='bridgeUID')
    configuration: Dict[str, Any] = {}
    properties: Dict[str, str] = {}
    uid: str = field(name='UID')
    thing_type: str = field(name='thingTypeUID')
    location: str = ''

    # EnrichedThingDTO
    # https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core.io.rest.core/src/main/java/org/openhab/core/io/rest/core/thing/EnrichedThingDTO.java
    channels: List[ChannelResp] = []
    status: ThingStatusResp = field(name='statusInfo')
    firmware_status: Optional[FirmwareStatusResp] = None
    editable: bool



import random
import string
from pathlib import Path
from typing import Literal
from typing import Optional, Tuple

import pydantic
from pydantic import Field

from easyconfig.models import BaseModel

QOS = Literal[0, 1, 2]


class TLSSettings(BaseModel):
    enabled: bool = Field(default=True, description='Enable TLS for the connection')
    ca_cert: Path = Field(
        default='', description='Path to a CA certificate that will be treated as trusted', alias='ca cert')
    insecure: bool = Field(
        default=False, description='Validate server hostname in server certificate')


class Connection(BaseModel):
    client_id: str = Field('HABApp-' + ''.join(random.choices(string.ascii_letters, k=13)),
                           description='ClientId that is used to uniquely identify this client on the mqtt broker.')
    host: str = Field('', description='Connect to this host. Empty string ("") disables the connection.')
    port: int = 1883
    user: str = ''
    password: str = ''
    tls: TLSSettings = Field(default_factory=TLSSettings)


class Subscribe(BaseModel):
    qos: QOS = Field(default=0, description='Default QoS for subscribing')
    topics: Tuple[Tuple[str, Optional[QOS]], ...] = Field(default=('#', ))

    @pydantic.field_validator('topics', mode='before')
    def parse_topics(cls, v):
        if not isinstance(v, (list, tuple, set)):
            raise ValueError('must be a list')
        ret = []
        for e in v:
            if isinstance(e, str):
                e = (e, None)
            ret.append(tuple(e))
        return tuple(ret)


class Publish(BaseModel):
    qos: QOS = Field(default=0, description='Default QoS when publishing values')
    retain: bool = Field(default=False, description='Default retain flag when publishing values')


class General(BaseModel):
    listen_only: bool = Field(False, description='If True HABApp does not publish any value to the broker')


class MqttConfig(BaseModel):
    """MQTT configuration"""

    connection: Connection = Field(default_factory=Connection)
    subscribe: Subscribe = Field(default_factory=Subscribe)
    publish: Publish = Field(default_factory=Publish)
    general: General = Field(default_factory=General)



from statistics import mean, median
from typing import Union


def format_duration(duration: Union[None, str, float]) -> str:
    if duration is None:
        return ' ' * 6
    if isinstance(duration, str):
        return f'{duration:^6s}'

    if duration < 0.0001:
        # 99.9ns
        return f'{duration * 1000 * 1000:4.1f}us'
    elif duration < 0.01:
        # 9.99ms
        return f'{duration * 1000:4.2f}ms'
    elif duration < 0.1:
        # 99.9ms
        return f'{duration * 1000:4.1f}ms'
    elif duration < 10:
        # 1.234s
        return f'{duration:5.3f}s'
    else:
        #  19.2s
        return f'{duration:5.1f}s'


class BenchContainer:
    def __init__(self):
        self.times = []

    def create(self, name: str) -> 'BenchTime':
        c = BenchTime(name)
        self.times.append(c)
        return c

    def show(self):
        indent = max(map(lambda x: len(x.name), self.times), default=0)
        BenchTime.show_table(indent)
        for b in self.times:
            b.show(indent)
        print('')


class BenchTime:

    @classmethod
    def show_table(cls, indent_name=0):
        print(f'{"":{indent_name}s} | {format_duration("dur")} | {"per sec":7s} | 'f'{format_duration("median")} | '
              f'{format_duration("min")} | {format_duration("max")} | {format_duration("mean")}')

    def __init__(self, name: str, factor: int = 1):
        self.name = name
        self.times = []
        self.factor = factor

    def show(self, indent_name=0):
        total = sum(self.times)
        count = len(self.times)
        _mean = mean(self.times) if self.times else 0
        _medi = median(self.times) if self.times else 0

        per_sec = ((count * self.factor) / total) if total else 0
        unit = ''
        if per_sec > 1000:
            per_sec /= 1000
            unit = 'k'
        if per_sec > 1000:
            per_sec /= 1000
            unit = 'm'

        print(f'{self.name:>{indent_name}s} | {format_duration(total)} | '
              f'{per_sec:{7 - len(unit)}.{3 - len(unit)}f}{unit} | '
              f'{format_duration(_medi)} | {format_duration(min(self.times, default=0))} | '
              f'{format_duration(max(self.times, default=0))} | '
              f'{format_duration(_mean)}')



from __future__ import annotations

import logging
from typing import Final

from HABApp.config import CONFIG
from HABApp.core.connections import BaseConnectionPlugin
from HABApp.core.internals import uses_item_registry
from HABApp.core.logger import log_warning
from HABApp.openhab.connection.connection import OpenhabConnection
from HABApp.openhab.connection.handler.func_async import async_get_things, async_get_links

PING_CONFIG: Final = CONFIG.openhab.ping

Items = uses_item_registry()


class BrokenLinksPlugin(BaseConnectionPlugin[OpenhabConnection]):

    def __init__(self, name: str | None = None):
        super().__init__(name)
        self.do_run = True

    async def on_online(self):
        if not self.do_run:
            return None
        self.do_run = False

        log = logging.getLogger('HABApp.openhab.links')

        things = await async_get_things()
        links = await async_get_links()

        available_things = {t.uid for t in things}
        available_channels = {c.uid for t in things for c in t.channels}

        for link in sorted(links, key=lambda x: x.channel):
            if not Items.item_exists(link.item):
                log_warning(log, f'Item "{link.item}" does not exist! '
                                 f'(link between item "{link.item:s}" and channel "{link.channel:s}")')
                continue

            if link.channel not in available_channels:
                # check if the thing exists
                thing_uid, channel_id = link.channel.rsplit(':', maxsplit=1)
                if thing_uid in available_things:
                    log_warning(log, f'Channel "{channel_id}" on thing "{thing_uid:s}" does not exist! '
                                     f'(link between item "{link.item:s}" and channel "{link.channel:s}")')
                else:
                    log_warning(log, f'Thing "{thing_uid:s}" does not exist! '
                                     f'(link between item "{link.item:s}" and channel "{link.channel:s}")')



from typing import Final, Tuple


TOPIC_INFOS: Final    = 'HABApp.Infos'
TOPIC_WARNINGS: Final = 'HABApp.Warnings'
TOPIC_ERRORS: Final   = 'HABApp.Errors'

TOPIC_FILES: Final = 'HABApp.Files'
TOPIC_CONNECTIONS: Final = 'HABApp.Connections'


ALL_TOPICS: Tuple[str, ...] = (
    TOPIC_INFOS, TOPIC_WARNINGS, TOPIC_ERRORS,

    TOPIC_FILES,
    TOPIC_CONNECTIONS
)
